@@ -421,7 +421,7 @@ MD5::checkFileChecksum(char const* const checksum,
 	std::string actual_checksum = getFileChecksum(filename, up_to_size);
 	result = (checksum == actual_checksum);
     }
-    catch (std::runtime_error)
+    catch (std::runtime_error const&)
     {
 	// Ignore -- return false
     }
