@@ -15,7 +15,7 @@ int main()
 	bool done = false;
 	while (! done)
 	{
-	    int len = fread(buf, 1, sizeof(buf), stdin);
+	    size_t len = fread(buf, 1, sizeof(buf), stdin);
 	    if (len <= 0)
 	    {
 		done = true;
